import pathlib
import tempfile
import unittest
import unittest.mock

from irisml.tasks.upload_azure_blob_directory import Task


class TestUploadAzureBlobDirectory(unittest.TestCase):
    def test_simple(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            (temp_dir / 'file1').write_text('file1')
            (temp_dir / 'file2').write_text('file2')
            (temp_dir / 'dir1').mkdir()
            (temp_dir / 'dir1' / 'file3').write_text('file3')
            (temp_dir / 'dir1' / 'dir2').mkdir()
            (temp_dir / 'dir1' / 'dir2' / 'file4').write_text('file4')
            with unittest.mock.patch('irisml.tasks.upload_azure_blob_directory.ContainerClient') as m_client:
                Task(Task.Config('https://fake_url/', temp_dir)).execute(Task.Inputs())
                m_client_instance = m_client.from_container_url.return_value
                self.assertEqual(m_client_instance.upload_blob.call_count, 4)



import unittest
import pickle
import pathlib
import tempfile

import torch
from irisml.tasks.load_tensor_list import Task


class TestLoadTensorList(unittest.TestCase):
    def test_tensor_list(self):
        tensor_list = [torch.Tensor([0, 1, 0, 0, 1, 1]), torch.Tensor([[1, 1, 0, 0, 1, 1]])]
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'test.pth'
            temp_file.write_bytes(pickle.dumps(tensor_list))
            data = Task(Task.Config()).execute(Task.Inputs(path=temp_file)).tensor_list
            self.assertTrue(all([torch.equal(a, b) for a, b in zip(data, tensor_list)]))



import unittest
import PIL.Image
import torch
from irisml.tasks.extract_image_bytes_from_dataset import Task


class TestExtractImageBytesFromDataset(unittest.TestCase):
    def test_simple(self):
        fake_image = PIL.Image.new('RGB', (300, 300))
        dataset = [(fake_image, torch.tensor(0))] * 5 + [(fake_image, torch.tensor(1))] * 5

        outputs = Task(Task.Config(num_images_per_class=1)).execute(Task.Inputs(dataset=dataset, num_classes=2))
        self.assertEqual(len(outputs.image_bytes), 2)
        self.assertTrue(isinstance(b, bytes) for b in outputs.image_bytes)

        outputs = Task(Task.Config(num_images_per_class=10)).execute(Task.Inputs(dataset=dataset, num_classes=2))
        self.assertEqual(len(outputs.image_bytes), 20)
        self.assertTrue(isinstance(b, bytes) for b in outputs.image_bytes)

        outputs = Task(Task.Config(num_images_per_class=10)).execute(Task.Inputs(dataset=dataset, num_classes=3))
        self.assertEqual(len(outputs.image_bytes), 30)
        self.assertTrue(isinstance(b, bytes) for b in outputs.image_bytes)

    def test_dry_run(self):
        outputs = Task(Task.Config(num_images_per_class=1)).dry_run(Task.Inputs(dataset=[], num_classes=2))
        self.assertEqual(len(outputs.image_bytes), 2)
        self.assertTrue(isinstance(b, bytes) for b in outputs.image_bytes)

        outputs = Task(Task.Config(num_images_per_class=10)).dry_run(Task.Inputs(dataset=[], num_classes=2))
        self.assertEqual(len(outputs.image_bytes), 20)
        self.assertTrue(isinstance(b, bytes) for b in outputs.image_bytes)

        outputs = Task(Task.Config(num_images_per_class=10)).dry_run(Task.Inputs(dataset=[], num_classes=3))
        self.assertEqual(len(outputs.image_bytes), 30)
        self.assertTrue(isinstance(b, bytes) for b in outputs.image_bytes)



import json
import pathlib
import tempfile
import unittest
from irisml.tasks.load_strs_from_json_file import Task


class TestLoadStrsFromJsonFile(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            filepath = pathlib.Path(f.name)

            # List[str]
            filepath.write_text(json.dumps(["a", "b", "c"]))
            outputs = Task(Task.Config(filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, ["a", "b", "c"])

            # Dict[int, str]
            filepath.write_text(json.dumps({0: "a", 2: "c", 1: "b"}))
            outputs = Task(Task.Config(filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, ["a", "b", "c"])

            # Missing key
            filepath.write_text(json.dumps({0: "a", 2: "c"}))
            outputs = Task(Task.Config(filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, ["a", "", "c"])



import dataclasses
import logging
import irisml.core


logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Join two strings to one string."""
    VERSION = '0.0.1'

    @dataclasses.dataclass
    class Inputs:
        string1: str = ''
        string2: str = ''

    @dataclasses.dataclass
    class Config:
        delimiter: str = ' '

    @dataclasses.dataclass
    class Outputs:
        string: str

    def execute(self, inputs):
        if not inputs.string1 and not inputs.string2:
            raise ValueError("string1 or string2 must be non-empty!")
        elif not inputs.string1 or not inputs.string2:
            return self.Outputs(inputs.string1 + inputs.string2)
        else:
            string = inputs.string1 + self.config.delimiter + inputs.string2
            logging.info(f'Joined string: {string}')
            return self.Outputs(string)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import typing
import irisml.core


PLACEHOLDER = '<|placeholder|>'
INDEX_PLACEHOLDER = '<|index|>'
logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Make a prompt with a list of strings.

    For example, if the template is "What is <placeholder>?" and the strings are ["a", "b", "c"], the prompt will be "What is a, b, c?".

    Optionally, you can specify a sub_template to use for each string. For example, if the template is "What is <|placeholder|>?" and
    the sub_template is "<|index|>.<|placeholder|>", the prompt will be "What is 0.a, 1.b, 2.c?".

    Config:
        template (str): The template to use for the prompt. Must contain "<|placeholder|>".
        sub_template (str): The template to use for each string. Must contain "<|placeholder|>". Can also contain "<|index|>".
        delimiter (str): The delimiter to use between the strings. Defaults to ", ".
    """
    VERSION = '0.2.0'

    @dataclasses.dataclass
    class Inputs:
        strings: typing.List[str]

    @dataclasses.dataclass
    class Config:
        template: str
        sub_template: str = PLACEHOLDER
        delimiter: str = ', '

    @dataclasses.dataclass
    class Outputs:
        prompt: str

    def execute(self, inputs):
        if PLACEHOLDER not in self.config.template:
            raise ValueError(f'"{PLACEHOLDER}" must be in template')
        if PLACEHOLDER not in self.config.sub_template:
            raise ValueError(f'"{PLACEHOLDER}" must be in sub_template')

        strings = [self.config.sub_template.replace(PLACEHOLDER, s).replace(INDEX_PLACEHOLDER, str(i)) for i, s in enumerate(inputs.strings)]
        prompt = self.config.template.replace(PLACEHOLDER, self.config.delimiter.join(strings))
        logger.info(f"Created a prompt: {prompt}")
        return self.Outputs(prompt=prompt)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
from irisml.tasks.make_prompt_with_strings import Task


class TestMakePromptWithStrings(unittest.TestCase):
    def test_simple(self):
        output = Task(Task.Config(template='What is <|placeholder|>?')).execute(Task.Inputs(strings=['a', 'b', 'c']))
        self.assertEqual(output.prompt, 'What is a, b, c?')

    def test_sub_template(self):
        output = Task(Task.Config(template='Here is a list. <|placeholder|>', sub_template='<|index|>: <|placeholder|>')).execute(Task.Inputs(strings=['a', 'b', 'c']))
        self.assertEqual(output.prompt, 'Here is a list. 0: a, 1: b, 2: c')



import io
import unittest
import torch
from irisml.tasks.serialize_tensor import Task


class TestSerializeTensor(unittest.TestCase):
    def test_simple(self):
        tensor = torch.rand(3, 4, 5)
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor))
        self.assertTrue(torch.equal(torch.load(io.BytesIO(outputs.data)), tensor))



import json
import unittest
from irisml.tasks.get_int_from_json_strings import Task


class TestGetIntFromJsonStrings(unittest.TestCase):
    def test_simple(self):
        json_strings = [json.dumps({'a': 3, 'b': 1}), json.dumps({'a': 2, 'b': 2, 'c': 3})]
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3, 2])
        outputs = Task(Task.Config(key_name='b')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [1, 2])
        outputs = Task(Task.Config(key_name='c')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [-1, 3])

    def test_voting(self):
        json_strings = ['{"a": 3}{"a": 4}{"a": 4}"']
        outputs = Task(Task.Config(key_name='a', voting=True)).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [4])

    def test_noisy_json(self):
        json_strings = ['output: {"a": 3,\n "b": 1}.\n', 'output: {"a": 2, "b": 2, "c": {"d": 4}}.']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3, 2])
        outputs = Task(Task.Config(key_name='b')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [1, 2])
        outputs = Task(Task.Config(key_name='c')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [-1, -1])

    def test_multiple_json(self):
        json_strings = ['{"a": 3,\n"b": 1}{"a": 4, \n"b": 2}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])

    def test_json_with_single_quotes(self):
        json_strings = ["{'a': 3, 'b': 1}", "{'a': 2, 'b': 2, 'c': 3}"]
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3, 2])

    def test_invalid_json(self):
        json_strings = ['{"a": 3,', '{"b": "abc"!?aksjhdf38totallyinvalid-=+"""a": 8', '{"a":      -100', '{"a": "abc']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3, 8, -100, -1])

    def test_no_json(self):
        json_strings = ['this is just a string']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [-1])

    def test_json_with_unescaped_double_quotes(self):
        json_strings = ['{"a": 3, "b": "Includes "double quotes""}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])

        json_strings = ['{"a": 3, "b": "Includes "double quotes" and some strings"}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])

        # Has a comma right after the double quotes
        json_strings = ['random "string", then {"a": 3, "b": "Includes "double quotes", another string"}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])

        json_strings = ['{"a": 3, "b": "Includes a single " double quotes"}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])

        json_strings = ['{"a": 3, "b": "Includes many """""""" double """""" quotes"}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])

        json_strings = ['{"a": 3, "b": "Includes escaped \\" double quotes"}']
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertEqual(outputs.tensor.tolist(), [3])



import dataclasses
import logging
import pathlib
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save the model's state_dict to the specified file.

    Config:
        path (pathlib.Path): The output filepath.
        fp16 (bool): If True, FP32 tensors will be converted to FP16.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path
        fp16: bool = False

    def execute(self, inputs):
        state_dict = inputs.model.state_dict()

        if self.config.fp16:
            original_size = self._get_size_in_bytes(state_dict)
            state_dict = {k: (v.half() if torch.is_tensor(v) and v.dtype == torch.float32 else v) for k, v in state_dict.items()}
            new_size = self._get_size_in_bytes(state_dict)
            logger.info(f"Converted the state_dict to FP16. {original_size=}, {new_size=}")

        self.config.path.parent.mkdir(exist_ok=True, parents=True)
        torch.save(state_dict, self.config.path)
        return self.Outputs()

    @staticmethod
    def _get_size_in_bytes(state_dict):
        return sum(t.element_size() * t.numel() for t in state_dict.values() if torch.is_tensor(t))



import dataclasses
import io
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Serialize a pytorch tensor."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        tensor: torch.Tensor

    @dataclasses.dataclass
    class Outputs:
        data: bytes

    def execute(self, inputs):
        bytes_io = io.BytesIO()
        torch.save(inputs.tensor, bytes_io)
        data = bytes_io.getvalue()
        logger.info(f"Serialized a tensor (shape: {inputs.tensor.shape}). {len(data)} bytes.")
        return self.Outputs(data)

    def dry_run(self, inputs):
        return self.execute(inputs)



import collections
import typing
import unittest
import torch
from irisml.tasks.get_dataset_split import Task


class DummyDataset(torch.utils.data.Dataset):
    def __init__(self, data: typing.List):
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, x):
        return None, self._data[x]


class DummyDatasetWithGetTargets(DummyDataset):
    def get_targets(self, index):
        return self._data[index]

    def __getitem__(self, x):
        raise RuntimeError("get_targets() should be used instead.")


class TestGetDatasetSubset(unittest.TestCase):
    def test_simple(self):
        dataset = DummyDataset(list(range(100)))

        inputs = Task.Inputs(train_dataset=dataset)
        config = Task.Config(train_val_split=0.8)

        outputs = Task(config).execute(inputs)
        self.assertEqual(len(outputs.train_dataset), 80)
        self.assertEqual(len(outputs.val_dataset), 20)

    def test_with_get_targets(self):
        # The dataset has 'get_targets" method.
        dataset = DummyDatasetWithGetTargets([torch.tensor(i // 10) for i in range(100)])
        outputs = Task(Task.Config(train_val_split=0.8, keep_class_distribution=True)).execute(Task.Inputs(dataset))

        self.assertEqual(len(outputs.train_dataset), 80)
        self.assertEqual(len(outputs.val_dataset), 20)

    def test_edge(self):
        dataset = DummyDataset(list(range(100)))

        inputs = Task.Inputs(train_dataset=dataset)
        config = Task.Config(train_val_split=0.0)

        outputs = Task(config).execute(inputs)
        self.assertEqual(len(outputs.train_dataset), 0)
        self.assertEqual(len(outputs.val_dataset), 100)

    def test_edge2(self):
        dataset = DummyDataset(list(range(100)))

        inputs = Task.Inputs(train_dataset=dataset)
        config = Task.Config(train_val_split=1.0)

        outputs = Task(config).execute(inputs)
        self.assertEqual(len(outputs.train_dataset), 100)
        self.assertEqual(len(outputs.val_dataset), 0)

    def test_keep_distribution(self):
        dataset = DummyDataset([torch.tensor(i // 10) for i in range(100)])

        outputs = Task(Task.Config(train_val_split=0.8, keep_class_distribution=True)).execute(Task.Inputs(dataset))
        class_counter = collections.Counter([int(x[1]) for x in outputs.train_dataset])
        self.assertEqual(list(class_counter.values()), [8] * 10)

        class_counter = collections.Counter([int(x[1]) for x in outputs.val_dataset])
        self.assertEqual(list(class_counter.values()), [2] * 10)

    def test_keep_distribution_train_prioritized(self):
        # Train classes should have at least one image.
        dataset = DummyDataset([torch.tensor(i) for i in range(100)])

        outputs = Task(Task.Config(train_val_split=0.8, keep_class_distribution=True)).execute(Task.Inputs(dataset))
        # Since each class has only 1 sample, all samples will be used in train dataset.
        self.assertEqual(len(outputs.train_dataset), 100)
        self.assertEqual(len(outputs.val_dataset), 0)

    def test_keep_distribution_od(self):
        dataset = DummyDataset([torch.tensor([[i // 10, 0, 0, 1, 1]]) for i in range(100)])

        outputs = Task(Task.Config(train_val_split=0.8, keep_class_distribution=True)).execute(Task.Inputs(dataset))
        class_counter = collections.Counter([int(x[1][0][0]) for x in outputs.train_dataset])
        self.assertEqual(list(class_counter.values()), [8] * 10)

    def test_keep_distribution_with_negative_samples(self):
        dataset = DummyDataset([torch.empty(0) for i in range(10)] + [torch.tensor(i // 10) for i in range(10, 100)])

        outputs = Task(Task.Config(train_val_split=0.8, keep_class_distribution=True)).execute(Task.Inputs(dataset))
        class_counter = collections.Counter([int(x[1]) for x in outputs.train_dataset if not x[1].shape or len(x[1])])
        self.assertEqual(list(class_counter.values()), [8] * 9)
        self.assertEqual(sum(1 for x, c in outputs.train_dataset if c.shape and len(c) == 0), 8)  # The number of negative samples.

        class_counter = collections.Counter([int(x[1]) for x in outputs.val_dataset if not x[1].shape or len(x[1])])
        self.assertEqual(list(class_counter.values()), [2] * 9)
        self.assertEqual(sum(1 for x, c in outputs.val_dataset if c.shape and len(c) == 0), 2)  # The number of negative samples.

    def test_keep_distribution_multilabel(self):
        dataset = DummyDataset([torch.tensor([i // 10, i % 2 + 10]) for i in range(100)])

        outputs = Task(Task.Config(train_val_split=0.8, keep_class_distribution=True)).execute(Task.Inputs(dataset))
        num_ten_class = sum(1 for x in outputs.train_dataset if 10 in x[1].tolist())
        self.assertGreater(num_ten_class, 30)

        num_ten_class = sum(1 for x in outputs.val_dataset if 10 in x[1].tolist())
        self.assertLess(num_ten_class, 20)



import dataclasses
import logging
import typing
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobClient, ContainerClient
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Upload a binary file to Azure Storage Blob.

    For the details of Config, see download_azure_blob task's help.

    Raises:
        azure.storage.exceptions if failed.

    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        data: bytes

    @dataclasses.dataclass
    class Config:
        blob_url: typing.Optional[str] = None
        container_url: typing.Optional[str] = None
        blob_path: typing.Optional[str] = None

    def execute(self, inputs):
        if self.config.blob_url and (self.config.container_url or self.config.blob_path):
            raise ValueError("You cannot specify both blob_url and container_url.")

        if not (self.config.blob_url or self.config.container_url):
            raise ValueError("You must specify either blob_url or container_url.")

        if self.config.container_url and not self.config.blob_path:
            raise ValueError("You must specify blob_path if container_url is used.")

        if self.config.blob_url:
            blob_client = BlobClient.from_blob_url(self.config.blob_url, credential=DefaultAzureCredential())
            blob_client.upload_blob(inputs.data)
        elif self.config.container_url:
            container_client = ContainerClient.from_container_url(self.config.container_url, credential=DefaultAzureCredential())
            container_client.upload_blob(self.config.blob_path, inputs.data, max_concurrency=8, timeout=300)

        logger.info(f"Uploaded {len(inputs.data)} bytes.")
        return self.Outputs()



import cProfile
import dataclasses
import logging
import pathlib
import tempfile
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Run profiler on the given tasks.

    cProfile is used to profile the task.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False
    RESOLVE_CONFIG_VARIABLES = False

    @dataclasses.dataclass
    class Config:
        tasks: typing.List[irisml.core.TaskDescription]

    @dataclasses.dataclass
    class Outputs:
        stats_bytes: bytes

    def execute(self, inputs):
        tasks = [irisml.core.Task(t) for t in self.config.tasks]

        logger.info(f"Starting profiling for {len(tasks)} tasks.")
        with cProfile.Profile() as p:
            for task in tasks:
                task.load_module()
                task.execute(self.context)

        logger.info("Profiling finished. Printing stats.")

        p.print_stats()
        with tempfile.NamedTemporaryFile() as f:
            p.dump_stats(f.name)
            stats_bytes = pathlib.Path(f.name).read_bytes()

        return self.Outputs(stats_bytes=stats_bytes)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
from azure.keyvault.secrets import SecretClient
from azure.identity import DefaultAzureCredential
import irisml.core


class Task(irisml.core.TaskBase):
    """Get a secret from Azure KeyVault."""
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        keyvault_name: str
        secret_name: str

    @dataclasses.dataclass
    class Outputs:
        secret: str = ''

    def execute(self, inputs):
        vault_url = f'https://{self.config.keyvault_name}.vault.azure.net'
        client = SecretClient(vault_url=vault_url, credential=DefaultAzureCredential())

        secret = client.get_secret(self.config.secret_name)
        assert isinstance(secret.value, str), f"Unexpected secret type: {type(secret.value)}"
        return self.Outputs(secret.value)



import dataclasses
import time
import irisml.core


class Task(irisml.core.TaskBase):
    """Get the current time in seconds since the epoch"""
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Outputs:
        time: float

    def execute(self, inputs):
        return self.Outputs(time.time())

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import json
import logging
import re
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get a list of ints from a JSON string.

    This task takes a list of JSON strings and returns a list of lists of ints.
    Each JSON string must contain a key with the name specified in the
    configuration. The value of this key must be a list of ints.

    Config:
        key_name (str): The name of the key in the JSON string that contains the list of ints.
    """
    VERSION = '1.0.1'

    @dataclasses.dataclass
    class Inputs:
        json_strings: typing.List[str]

    @dataclasses.dataclass
    class Config:
        key_name: str

    @dataclasses.dataclass
    class Outputs:
        tensors: typing.List[torch.Tensor]

    def execute(self, inputs):
        tensors = []
        for i, json_string in enumerate(inputs.json_strings):
            values = []
            try:
                values = self._get_ints_from_json(json_string)
                logger.info(f"Index {i}: Found values {values} in json: {repr(json_string)}")
            except Exception as e:
                logger.warning(f"Index {i}: {repr(e)}. json: {repr(json_string)}")

            tensors.append(torch.tensor(values, dtype=torch.int64))

        assert len(tensors) == len(inputs.json_strings)
        return self.Outputs(tensors)

    def dry_run(self, inputs):
        return self.execute(inputs)

    def _get_ints_from_json(self, json_str):
        json_dict, _ = self._decode_json(json_str)
        if json_dict is None:
            raise ValueError("JSON string could not be decoded.")

        if self.config.key_name not in json_dict:
            raise ValueError(f"Key {self.config.key_name} not found in JSON string.")

        result = json_dict[self.config.key_name]
        if not isinstance(result, list) or not all(isinstance(x, int) for x in result):
            raise ValueError(f"Key {self.config.key_name} is not a list of ints.")

        return result

    @staticmethod
    def _decode_json(json_str):
        """Decode a JSON string.

        This method tries to decode a JSON string. If the string cannot be
        decoded, it tries to fix the string and decode it again. If the string
        still cannot be decoded, it returns None.

        Returns:
            A tuple of the decoded JSON object and the index of the end of the
            JSON string in the original string.
        """
        start = json_str.find('{')
        if start == -1:
            return None, len(json_str)

        try:
            return json.JSONDecoder().raw_decode(json_str[start:])
        except json.JSONDecodeError:
            pass

        try:
            # Try replacing a single quote with a double quote and vice versa.
            trans = str.maketrans('\'"', '"\'')
            return json.JSONDecoder().raw_decode(json_str.translate(trans)[start:])
        except json.JSONDecodeError:
            pass

        try:
            # Try replacing a single quote with a double quote
            return json.JSONDecoder().raw_decode(json_str.replace('\'', '"')[start:])
        except json.JSONDecodeError:
            pass

        try:
            # Try escaping a double quote in the JSON string
            pattern = r'([:\[,{]\s*)"(.*?)"(?=\s*[:,\]}])'

            def replace(match):
                return match.group(1) + '"' + match.group(2).replace('"', "'") + '"'
            fixed_json = re.sub(pattern, replace, json_str[start:])
            return json.JSONDecoder().raw_decode(fixed_json)
        except json.JSONDecodeError:
            pass

        return None, len(json_str)



import dataclasses
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get the largest Topk values and indices.

    Inputs:
        tensor (torch.Tensor): The input tensor

    Config:
        k (int): The "k".
        device (str): The device to use. If not specified, it uses cuda if available.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        tensor: torch.Tensor

    @dataclasses.dataclass
    class Config:
        k: int
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None

    @dataclasses.dataclass
    class Outputs:
        values: torch.Tensor
        indices: torch.LongTensor

    def execute(self, inputs):
        device = self._get_device()
        result = torch.topk(inputs.tensor.to(device), self.config.k)
        return self.Outputs(result.values.to('cpu'), result.indices.to('cpu'))

    def dry_run(self, inputs):
        shape = [*inputs.tensor.shape[:-1], self.config.k]
        fake_values = torch.zeros(shape)
        fake_indices = torch.arange(self.config.k).expand(shape)
        return self.Outputs(fake_values, fake_indices)

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)



import pathlib
import tempfile
import unittest
from irisml.tasks.read_file import Task


class TestReadFile(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b'Hello World!')
            f.flush()
            outputs = Task(Task.Config(path=pathlib.Path(f.name))).execute(Task.Inputs())
            self.assertEqual(outputs.data, b'Hello World!')



import dataclasses
import difflib
import logging
import random
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Assigns a class to a string based on the class name being present in the string.

    This task uses multiple matchers to assign a class to a string. The matchers are
    executed in the following order:

    1. ExactClassMatcher: Find all class names that are present in the string, assigns the longest class name
    to the string.
    2. WordCountClassMatcher: Count the number of matching words between the class name
    and the string. If the number of matching words is greater than 0, it assigns the
    class to the string.
    3. DifflibClassMatcher: Uses Difflib to find the class name that is most similar to the string.

    Config:
        assign_negative_class_if_no_match: If True, assigns -1 to the string if it doesn't find a match by ExactClassMatcher.
    """
    VERSION = '0.1.4'

    @dataclasses.dataclass
    class Inputs:
        strings: typing.List[str]
        class_names: typing.List[str]

    @dataclasses.dataclass
    class Config:
        assign_negative_class_if_no_match: bool = False

    @dataclasses.dataclass
    class Outputs:
        classes: torch.Tensor  # Shape: (N, )

    def execute(self, inputs):
        matchers = [ExactClassMatcher(inputs.class_names), WordCountClassMatcher(inputs.class_names), DifflibClassMatcher(inputs.class_names)]

        results = [None] * len(inputs.strings)
        for i, matcher in enumerate(matchers):
            num_unknowns_before = results.count(None)
            results = [c if c is not None else matcher(s) for s, c in zip(inputs.strings, results)]
            num_unknowns_after = results.count(None)
            logger.info(f"Assigned {num_unknowns_before - num_unknowns_after}/{num_unknowns_before} strings based on {i}th matcher.")

            if num_unknowns_after == 0:
                break

            if i == 0:
                if self.config.assign_negative_class_if_no_match:
                    results = [c if c is not None else -1 for c in results]
                    logger.info(f"Assigned {num_unknowns_after}/{len(inputs.strings)} strings to class -1")

        return self.Outputs(torch.tensor(results))

    def dry_run(self, inputs):
        return self.execute(inputs)


class ClassMatcher:
    def __init__(self, class_names):
        self.class_names = [n.lower() for n in class_names]

    def __call__(self, s):
        return self.find(s.lower())

    def find(self, s):
        raise NotImplementedError


class ExactClassMatcher(ClassMatcher):
    def find(self, s):
        idx = -1
        max_length = 0
        for i, c in enumerate(self.class_names):
            if c in s and len(c) > max_length:
                max_length = len(c)
                idx = i
        return None if idx == -1 else idx


class WordCountClassMatcher(ClassMatcher):
    def __init__(self, class_names):
        super().__init__(class_names)
        self.class_name_words = [set(n.split()) for n in self.class_names]

    def find(self, s):
        word_set = set(s.split())
        matched_word_counts = [len(c.intersection(word_set)) / len(c) for c in self.class_name_words]
        max_matched_word_count = max(matched_word_counts)
        if max_matched_word_count == 0:
            return None
        max_indexes = [i for i, x in enumerate(matched_word_counts) if x == max_matched_word_count]
        class_index = random.choice(max_indexes) if len(max_indexes) > 1 else max_indexes[0]
        logger.debug(f"Assigned {repr(self.class_names[class_index])} to {repr(s)} by score {max_matched_word_count}")
        return class_index


class DifflibClassMatcher(ClassMatcher):
    def find(self, s):
        scores = [difflib.SequenceMatcher(None, s, c).ratio() for c in self.class_names]
        max_scores = max(scores)
        max_indexes = [i for i, x in enumerate(scores) if x == max_scores]
        class_index = random.choice(max_indexes) if len(max_indexes) > 1 else max_indexes[0]
        logger.debug(f"Assigned {repr(self.class_names[class_index])} to {repr(s)} by Difflib")
        return class_index



import dataclasses
import logging
import random
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get a subset of a dataset.

    Config:
        num_images (int): Number of images to keep in the dataset.
        verbose (bool): Shows a list of selected samples.
    """
    VERSION = '0.1.2'

    @dataclasses.dataclass
    class Config:
        num_images: int
        verbose: bool = False

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        num_images = len(inputs.dataset)
        new_num_images = self.config.num_images
        indexes = random.sample(range(num_images), new_num_images)
        if self.config.verbose:
            logger.info(f"Selected indexes: {indexes}")

        new_dataset = torch.utils.data.Subset(inputs.dataset, indexes)
        return self.Outputs(new_dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import io
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Deserialize a pytorch tensor."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        data: bytes

    @dataclasses.dataclass
    class Outputs:
        tensor: torch.Tensor

    def execute(self, inputs):
        tensor = torch.load(io.BytesIO(inputs.data))
        if not isinstance(tensor, torch.Tensor):
            raise ValueError(f"The deserialized object has unexpected type: {type(tensor)}")
        return self.Outputs(tensor)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
from typing import List, Optional
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """'If' conditional branch.

    If the condition is True, runs the tasks in 'then_tasks'. If the 'condition' is False, run the tasks in 'else_tasks'.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False
    RESOLVE_CONFIG_VARIABLES = False

    @dataclasses.dataclass
    class Config:
        condition: bool
        then_tasks: List[irisml.core.TaskDescription]
        else_tasks: Optional[List[irisml.core.TaskDescription]] = dataclasses.field(default_factory=list)

    def execute(self, inputs):
        return self._execute_core(inputs, dry_run=False)

    def dry_run(self, inputs):
        return self._execute_core(inputs, dry_run=True)

    def _execute_core(self, inputs, dry_run):
        resolved_condition = self.context.resolve(self.config.condition)
        logger.info(f"Condition is {resolved_condition}")

        tasks = self.config.then_tasks if resolved_condition else self.config.else_tasks
        for task_description in tasks:
            task = irisml.core.Task(task_description)
            task.load_module()
            task.execute(self.context, dry_run=dry_run)

        return self.Outputs()



import collections
import dataclasses
import io
import logging
import random
import typing
import PIL.Image
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Extract images from a dataset and convert them to bytes.

    For example, if you have a dataset with 10 classes and you want to extract 5 images per class, you will get 50 images.

    Config:
        num_images_per_class (int): Number of images to extract per class.
        image_size (int): Size of the extracted images.
        random_seed (int): Random seed to use when extracting images.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        num_classes: int

    @dataclasses.dataclass
    class Config:
        num_images_per_class: int = 1
        image_size: int = 224
        random_seed: int = 0

    @dataclasses.dataclass
    class Outputs:
        image_bytes: typing.List[bytes]

    def execute(self, inputs):
        extracted_images = collections.defaultdict(list)
        random_list = random.Random(self.config.random_seed).sample(list(range(len(inputs.dataset))), k=len(inputs.dataset))
        for i in random_list:
            image, targets = inputs.dataset[i]
            class_id = self._get_class(targets)
            if len(extracted_images[class_id]) < self.config.num_images_per_class:
                extracted_images[class_id].append(image.resize((self.config.image_size, self.config.image_size)))

        # Add empty images if we could not extract enough images for a class
        for i in range(inputs.num_classes):
            if len(extracted_images[i]) < self.config.num_images_per_class:
                logger.warning(f"Could not extract {self.config.num_images_per_class} images for class {i}. Actual: {len(extracted_images[i])}. Adding empty images.")
                extracted_images[i].extend([PIL.Image.new('RGB', (self.config.image_size, self.config.image_size))] * (self.config.num_images_per_class - len(extracted_images[i])))

        # Convert images to bytes
        image_bytes = []
        for i in range(inputs.num_classes):
            for image in extracted_images[i]:
                bytes_io = io.BytesIO()
                image.save(bytes_io, format='JPEG')
                image_bytes.append(bytes_io.getvalue())

        return self.Outputs(image_bytes=image_bytes)

    def dry_run(self, inputs):
        bytes_io = io.BytesIO()
        PIL.Image.new('RGB', (self.config.image_size, self.config.image_size)).save(bytes_io, format='JPEG')
        fake_image_bytes = bytes_io.getvalue()
        return self.Outputs(image_bytes=[fake_image_bytes] * inputs.num_classes * self.config.num_images_per_class)

    @staticmethod
    def _get_class(targets):
        if not isinstance(targets, torch.Tensor):
            raise ValueError(f"Expected targets to be a torch.Tensor, got {type(targets)}")
        return int(targets)



import contextlib
import io
import unittest
from irisml.tasks.print_environment_info import Task


class TestPrintEnvironmentInfo(unittest.TestCase):
    def test_print_something(self):
        with contextlib.redirect_stdout(io.StringIO()) as f:
            Task(Task.Config()).execute(Task.Inputs())
        s = f.getvalue()
        self.assertGreater(len(s), 0)



import pathlib
import tempfile
import unittest

import torch
from irisml.tasks.save_state_dict import Task


class TestLoadStateDict(unittest.TestCase):
    def test_simple(self):
        fake_model = torch.nn.Conv2d(3, 3, 3)

        inputs = Task.Inputs(model=fake_model)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'test.pth'

            Task(Task.Config(path=temp_file)).execute(inputs)
            self.assertTrue(self._is_equal_state_dict(fake_model.state_dict(), torch.load(temp_file)))

    def test_fp16(self):
        fake_model = torch.nn.Conv2d(3, 3, 3)

        inputs = Task.Inputs(model=fake_model)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'test.pth'

            Task(Task.Config(path=temp_file, fp16=True)).execute(inputs)
            saved = torch.load(temp_file)
            self.assertEqual(saved['weight'].dtype, torch.float16)
            self.assertEqual(saved['bias'].dtype, torch.float16)

    def test_no_parent_directory(self):
        fake_model = torch.nn.Conv2d(3, 3, 3)

        inputs = Task.Inputs(model=fake_model)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'new_directory' / 'another_new_directory' / 'test.pth'

            self.assertFalse(temp_file.exists())
            Task(Task.Config(path=temp_file)).execute(inputs)
            self.assertTrue(temp_file.exists())

    def _is_equal_state_dict(self, state_dict1, state_dict2):
        if set(state_dict1) != set(state_dict2):
            return False

        for key in state_dict1:
            if not torch.equal(state_dict1[key], state_dict2[key]):
                return False
        return True



import unittest
import torch
from irisml.tasks.assign_class_to_strings import Task


class TestAssignClassToStrings(unittest.TestCase):
    def test_simple(self):
        strings = ["white cat", "black cat", "red dog", "blue dog"]
        class_names = ["cat", "dog"]
        outputs = Task(Task.Config()).execute(Task.Inputs(strings, class_names))
        self.assertIsInstance(outputs.classes, torch.Tensor)
        self.assertEqual(outputs.classes.shape, (4,))
        self.assertEqual(outputs.classes.tolist(), [0, 0, 1, 1])

    def test_exact_matcher(self):
        strings = ["black cat", "cat"]
        class_names = ["cat", "black cat"]
        outputs = Task(Task.Config()).execute(Task.Inputs(strings, class_names))
        self.assertEqual(outputs.classes.tolist(), [1, 0])

    def test_assign_similar_class(self):
        strings = ["white cat", "black cat", "red d0g", "blue d0g", "unknown"]
        class_names = ["cat", "dog"]
        outputs = Task(Task.Config()).execute(Task.Inputs(strings, class_names))
        self.assertIsInstance(outputs.classes, torch.Tensor)
        self.assertEqual(outputs.classes.shape, (5,))
        self.assertEqual(outputs.classes.tolist()[:4], [0, 0, 1, 1])
        self.assertIn(outputs.classes[4], (0, 1))

    def test_unknown_class_negative(self):
        strings = ["white cat", "black cat", "red dog", "blue dog", "unknown"]
        class_names = ["cat", "dog"]
        outputs = Task(Task.Config(assign_negative_class_if_no_match=True)).execute(Task.Inputs(strings, class_names))
        self.assertIsInstance(outputs.classes, torch.Tensor)
        self.assertEqual(outputs.classes.shape, (5,))
        self.assertEqual(outputs.classes.tolist(), [0, 0, 1, 1, -1])



import dataclasses
from typing import Any, Optional
import irisml.core


class Task(irisml.core.TaskBase):
    """Assert the given input.

    For assert_equal and assert_not_equal, the supported type is str, int, float. List and Dict is also supported.
    For other assertions, only float is supported.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        value: Any

    @dataclasses.dataclass
    class Config:
        assert_equal: Optional[Any] = None
        assert_not_equal: Optional[Any] = None
        assert_almost_equal: Optional[float] = None
        assert_not_almost_equal: Optional[float] = None
        assert_greater: Optional[float] = None
        assert_greater_equal: Optional[float] = None
        assert_less: Optional[float] = None
        assert_less_equal: Optional[float] = None
        delta: float = 1e-2  # Threshold for assert_almost_equal and assert_not_almost_equal

    def execute(self, inputs):  # noqa: C901
        value = inputs.value

        if self.config.assert_equal is not None:
            if value != self.config.assert_equal:
                raise AssertionError(f"{value} != {self.config.assert_equal}")

        if self.config.assert_not_equal is not None:
            if value == self.config.assert_equal:
                raise AssertionError(f"{value} == {self.config.assert_equal}")

        if self.config.assert_almost_equal is not None:
            self._raise_if_not_number(value, "assert_almost_equal requires a number.")
            self._raise_if_not_number(self.config.assert_almost_equal, "assert_almost_equal requires a number.")

            diff = abs(value - self.config.assert_almost_equal)
            if diff > self.config.delta:
                raise AssertionError(f"Diff between {value} and {self.config.assert_almost_equal} is {diff} ( > {self.config.delta} )")

        if self.config.assert_not_almost_equal is not None:
            self._raise_if_not_number(value, "assert_not_almost_equal requires a number.")
            self._raise_if_not_number(self.config.assert_not_almost_equal, "assert_not_almost_equal requires a number.")

            diff = abs(value - self.config.assert_not_almost_equal)
            if diff < self.config.delta:
                raise AssertionError(f"Diff between {value} and {self.config.assert_not_almost_equal} is {diff} ( > {self.config.delta} )")

        if self.config.assert_greater is not None:
            self._raise_if_not_number(value, "assert_greater requires a number.")
            self._raise_if_not_number(self.config.assert_greater, "assert_greater requires a number.")

            if value <= self.config.assert_greater:
                raise AssertionError(f"{value} <= {self.config.assert_greater}")

        if self.config.assert_greater_equal is not None:
            self._raise_if_not_number(value, "assert_greater_equal requires a number.")
            self._raise_if_not_number(self.config.assert_greater_equal, "assert_greater_equal requires a number.")

            if value < self.config.assert_greater_equal:
                raise AssertionError(f"{value} < {self.config.assert_greater_equal}")

        if self.config.assert_less is not None:
            self._raise_if_not_number(value, "assert_less requires a number.")
            self._raise_if_not_number(self.config.assert_less, "assert_less requires a number.")

            if value >= self.config.assert_less:
                raise AssertionError(f"{value} >= {self.config.assert_less}")

        if self.config.assert_less_equal is not None:
            self._raise_if_not_number(value, "assert_less_equal requires a number.")
            self._raise_if_not_number(self.config.assert_less_equal, "assert_less_equal requires a number.")

            if value > self.config.assert_less_equal:
                raise AssertionError(f"{value} > {self.config.assert_less_equal}")

        return self.Outputs()

    def _raise_if_not_number(self, value, msg):
        if not isinstance(value, (int, float)):
            raise ValueError(f"{msg}: {type(value)}({value})")



import unittest
import torch
from irisml.tasks.print import Task


class TestPrint(unittest.TestCase):
    def test_float(self):
        with self.assertLogs(level='INFO') as cm:
            Task(Task.Config()).execute(Task.Inputs(data_float=0.123))
            self.assertEqual(cm.output, ['INFO:irisml.tasks.print:0.123'])

    def test_pp_float(self):
        with self.assertLogs(level='INFO') as cm:
            Task(Task.Config(pretty=True)).execute(Task.Inputs(data_float=0.123))
            self.assertEqual(cm.output, ['INFO:irisml.tasks.print:0.123'])

    def test_dict_str_float(self):
        with self.assertLogs(level='INFO') as cm:
            Task(Task.Config()).execute(Task.Inputs(data_dict_str_float={'a': 0.1, 'b': 0.2}))
            self.assertEqual(cm.output, ["INFO:irisml.tasks.print:{'a': 0.1, 'b': 0.2}"])

    def test_tensor(self):
        with self.assertLogs(level='INFO') as cm:
            Task(Task.Config()).execute(Task.Inputs(data_tensor=torch.zeros(2)))
            self.assertEqual(cm.output, ['INFO:irisml.tasks.print:Tensor shape=torch.Size([2]), dtype=torch.float32',
                                         'INFO:irisml.tasks.print:tensor([0., 0.])'])

    def test_list_str(self):
        with self.assertLogs(level='INFO') as cm:
            Task(Task.Config()).execute(Task.Inputs(data_list_str=['a', 'b']))
            self.assertEqual(cm.output, ['INFO:irisml.tasks.print:[\'a\', \'b\']'])

    def test_multiple_data(self):
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(data_int=3, data_float=0.1))

    def test_label(self):
        with self.assertLogs(level='INFO') as cm:
            Task(Task.Config('hello')).execute(Task.Inputs(data_float=0.123))
            self.assertEqual(cm.output, ['INFO:irisml.tasks.print:LABEL: hello', 'INFO:irisml.tasks.print:0.123'])



import dataclasses
import json
import logging
import pathlib
import typing
from collections import defaultdict
import torch
import irisml.core


logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load coco detections from a JSON to a list of tensors.

    The JSON file is a list of dictionaries, each dictionary should at least has key "image_id", "bbox" xywh normalized, "category_id". Optionally,
    the "score" field will also be loaded. All indices are supposed to start from 1.

    The output detections is a list of tensors, each tensor contains detections for an image, with shape (N, 5) or (N, 6), each row is [class (,score), x, y, x, y]. Absolute coordinates.
    If an image does not have prediction, N = 0.

    Config:
        filepath (Path): Path to the JSON file.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        filepath: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        detections: typing.List[torch.Tensor]

    def execute(self, inputs):
        data = json.loads(self.config.filepath.read_text())
        if not isinstance(data, list):
            raise TypeError('JSON file should contain a list!')
        has_score = 'score' in data[0]
        n_col_tensor = 6 if has_score else 5

        pred_by_image = defaultdict(list)
        for pred in data:
            box_xyxy = self._convert_xywh_to_xyxy(pred['bbox'])
            pred_by_image[pred["image_id"] - 1].append([pred["category_id"] - 1, pred["score"], *box_xyxy] if has_score else [pred["category_id"] - 1, *box_xyxy])

        detections = [torch.tensor(pred_by_image[img_id]) if img_id in pred_by_image else torch.empty((0, n_col_tensor)) for img_id in range(max(pred_by_image.keys())+1)]
        logger.info(f"Loaded {len(data)} detections with score? {has_score}")
        return self.Outputs(detections=detections)

    def _convert_xywh_to_xyxy(self, box):
        return [box[0], box[1], box[0] + box[2], box[1] + box[3]]

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import hashlib
import logging
import time
from typing import Optional
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobClient, ContainerClient
import tenacity
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Download a single blob from Azure Blob Storage.

    Returns the downloaded bytes.

    There are two ways to specify the blob.
    1. Use blob_url
    2. Use container_url and blob_path.

    If both urls are specified, raises ValueError.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        blob_url: Optional[str] = None
        container_url: Optional[str] = None
        blob_path: Optional[str] = None
        sha1_hash: Optional[str] = None

    @dataclasses.dataclass
    class Outputs:
        data: bytes = None

    def execute(self, inputs):
        if self.config.blob_url and (self.config.container_url or self.config.blob_path):
            raise ValueError("You cannot specify both blob_url and container_url.")

        if not (self.config.blob_url or self.config.container_url):
            raise ValueError("You must specify either blob_url or container_url.")

        if self.config.container_url and not self.config.blob_path:
            raise ValueError("You must specify blob_path if container_url is used.")

        start_time = time.time()
        downloaded_bytes = self._download()
        elapsed_time = time.time() - start_time
        file_hash = hashlib.sha1(downloaded_bytes).hexdigest()
        logger.info(f"Downloaded {len(downloaded_bytes)} bytes. SHA1 hash: {file_hash}, time: {elapsed_time:.2f}s")

        if self.config.sha1_hash and self.config.sha1_hash != file_hash:
            raise ValueError(f"The downloaded file has an unexpected hash value. Expected: {self.config.sha1_hash}. Actual: {file_hash}")

        return self.Outputs(downloaded_bytes)

    @tenacity.retry(reraise=True, stop=tenacity.stop_after_attempt(3))
    def _download(self):
        if self.config.blob_url:
            blob_client = BlobClient.from_blob_url(self.config.blob_url, credential=DefaultAzureCredential())
            downloaded_bytes = blob_client.download_blob().readall()
        elif self.config.container_url:
            container_client = ContainerClient.from_container_url(self.config.container_url, credential=DefaultAzureCredential())
            downloaded_bytes = container_client.download_blob(self.config.blob_path, max_concurrency=8, timeout=300).readall()

        return downloaded_bytes



import dataclasses
import logging
import random
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get train/test dataset for k-fold cross validation.

    Config:
        num_folds (int): K
        index (int): The index of a fold for test.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Config:
        num_folds: int
        index: int

    @dataclasses.dataclass
    class Outputs:
        train_dataset: torch.utils.data.Dataset
        val_dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        num_samples = len(inputs.dataset)

        if num_samples < self.config.num_folds:
            raise ValueError(f"There are not enough samples for k-fold cross validation. {num_samples=}, num_folds={self.config.num_folds}")

        if not 0 <= self.config.index < self.config.num_folds:
            raise ValueError(f"The index must be 0 <= index < num_folds. index={self.config.index}, num_folds={self.config.num_folds}")

        shuffled_indices = random.sample(range(num_samples), num_samples)

        fold_size = num_samples // self.config.num_folds
        remainder = num_samples % self.config.num_folds

        start_index = fold_size * self.config.index + min(self.config.index, remainder)
        end_index = start_index + fold_size + (1 if self.config.index < remainder else 0)
        assert 0 <= start_index < end_index <= num_samples

        train_indices = shuffled_indices[0:start_index] + shuffled_indices[end_index:]
        val_indices = shuffled_indices[start_index:end_index]

        assert len(train_indices) + len(val_indices) == num_samples

        train_dataset = torch.utils.data.Subset(inputs.dataset, train_indices)
        val_dataset = torch.utils.data.Subset(inputs.dataset, val_indices)
        logger.info(f"Created train/val datasets. {len(train_dataset)=}, {len(val_dataset)=}")

        return self.Outputs(train_dataset, val_dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Split string to a list of strings."""
    VERSION = '0.0.1'

    @dataclasses.dataclass
    class Inputs:
        string: str

    @dataclasses.dataclass
    class Config:
        delimiter: str = ','

    @dataclasses.dataclass
    class Outputs:
        strings: typing.List[str]

    def execute(self, inputs):
        return self.Outputs([s.strip() for s in inputs.string.split(self.config.delimiter)])



import unittest
from irisml.tasks.split_string import Task


class TestSplitString(unittest.TestCase):
    def test_simple(self):
        output = Task(Task.Config()).execute(Task.Inputs('a,b,c'))
        self.assertEqual(output.strings, ['a', 'b', 'c'])

        output = Task(Task.Config()).execute(Task.Inputs('a, b,   c'))
        self.assertEqual(output.strings, ['a', 'b', 'c'])

    def test_other_delimiter(self):
        output = Task(Task.Config(delimiter='-')).execute(Task.Inputs('a-b-c'))
        self.assertEqual(output.strings, ['a', 'b', 'c'])



import unittest
from irisml.tasks.make_random_choice_text_transform import Task


class TestMakeRandomChoiceTextTransform(unittest.TestCase):
    def test_simple(self):
        transform = Task(Task.Config()).execute(Task.Inputs()).transform

        self.assertEqual(transform('abc'), 'abc')
        self.assertIn(transform('a<|>b<|>c'), ['a', 'b', 'c'])

        all_results = set(transform('a<|>b<|>c') for _ in range(100))
        self.assertEqual(all_results, {'a', 'b', 'c'})



import unittest
from irisml.tasks.get_fake_object_detection_dataset import Task


class TestGetFakeObjectDetectionDataset(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(num_images=100, num_classes=10, num_max_boxes=10)).execute(None)
        dataset = outputs.dataset
        self.assertEqual(len(dataset), 100)
        self.assertEqual(len(outputs.class_names), 10)
        class_set = set(int(t[0]) for x in dataset for t in x[1])
        self.assertEqual(len(class_set), 10)
        self.assertTrue(all(len(x[1]) <= 10 for x in dataset))
        self.assertLess(sum(len(x[1]) for x in dataset), 1000)
        self.assertTrue(all(len(x[1].shape) == 2 for x in dataset))
        self.assertTrue(all(0 <= x[1] < x[3] <= 1 for _, t in dataset for x in t))
        self.assertTrue(all(0 <= x[2] < x[4] <= 1 for _, t in dataset for x in t))

    def test_multiple_image_sizes(self):
        outputs = Task(Task.Config(image_sizes=[(8, 8), (24, 24), (16, 16), (64, 2)])).execute(None)
        image_sizes_set = set(image.size for image, _ in outputs.dataset)
        self.assertEqual(image_sizes_set, {(8, 8), (24, 24), (16, 16), (64, 2)})

    def test_random_seed(self):
        outputs1 = Task(Task.Config(num_images=100, random_seed=0)).execute(Task.Inputs())
        outputs2 = Task(Task.Config(num_images=100, random_seed=0)).execute(Task.Inputs())

        targets1 = [d[1].tolist() for d in outputs1.dataset]
        targets2 = [d[1].tolist() for d in outputs2.dataset]
        self.assertEqual(targets1, targets2)

        outputs3 = Task(Task.Config(num_images=100, random_seed=1)).execute(Task.Inputs())
        targets3 = [d[1].tolist() for d in outputs3.dataset]
        self.assertNotEqual(targets1, targets3)

    def test_targets(self):
        outputs = Task(Task.Config(num_images=2, targets=[[[0, 0.25, 0.25, 0.5, 0.5], [1, 0.5, 0.5, 1.0, 1.0]], []])).execute(Task.Inputs())
        self.assertEqual(len(outputs.dataset), 2)
        self.assertEqual(outputs.dataset[0][1].tolist(), [[0, 0.25, 0.25, 0.5, 0.5], [1, 0.5, 0.5, 1.0, 1.0]])
        self.assertEqual(outputs.dataset[1][1].tolist(), [])

    def test_min_box_size(self):
        outputs = Task(Task.Config(image_sizes=[(4, 4)])).execute(Task.Inputs())
        for _, targets in outputs.dataset:
            for t in targets:
                # Since the image size is 4x4, the minimum box size is 0.25x0.25.
                self.assertGreaterEqual(t[3] - t[1], 0.25)
                self.assertGreaterEqual(t[4] - t[2], 0.25)



import dataclasses
import logging
import typing
import torch
import irisml.core


PLACEHOLDER = '<|placeholder|>'
INDEX_PLACEHOLDER = '<|index|>'
logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Make a list of prompts from a template and a list of strings.

    The template should contain exactly one instance of "<|placeholder|>".
    The sub_template should contain exactly one instance of "<|placeholder|>". It may also contain "<|index|>".

    `indices` is a list of lists of indices into `strings`. Each index list will be used to fill in the sub_template.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        strings: typing.List[str]
        indices: typing.List[torch.Tensor] = dataclasses.field(default_factory=list)

    @dataclasses.dataclass
    class Config:
        template: str
        sub_template: str = PLACEHOLDER
        delimiter: str = ', '

    @dataclasses.dataclass
    class Outputs:
        prompts: typing.List[str]

    def execute(self, inputs):
        if PLACEHOLDER not in self.config.template:
            raise ValueError(f'"{PLACEHOLDER}" not found in template')
        if PLACEHOLDER not in self.config.sub_template:
            raise ValueError(f'"{PLACEHOLDER}" not found in sub_template')
        if not all(isinstance(i, torch.Tensor) for i in inputs.indices):
            raise ValueError("indices must be a list of torch.Tensor")

        strings = [self.config.sub_template.replace(PLACEHOLDER, s).replace(INDEX_PLACEHOLDER, str(i)) for i, s in enumerate(inputs.strings)]

        prompts = []
        for index, indices in enumerate(inputs.indices):
            prompt = self.config.template.replace(PLACEHOLDER, self.config.delimiter.join([strings[i] for i in indices]))
            prompts.append(prompt)
            logger.info(f"Prompt {index}: {repr(prompt)}")

        return self.Outputs(prompts)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import pathlib
import irisml.core


class Task(irisml.core.TaskBase):
    """Join a given dir_path and a filename.

    Config:
        dir_path (pathlib.Path): Directory path.
        filename (str): A filename.

    Outputs:
        path (pathlib.Path): Joined filepath.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        dir_path: pathlib.Path
        filename: str

    @dataclasses.dataclass
    class Outputs:
        path: pathlib.Path

    def execute(self, inputs):
        return self.Outputs(self.config.dir_path / self.config.filename)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
from irisml.tasks.convert_string_to_string_list import Task


class TestConvertStringToStringList(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs(string='foo'))
        self.assertEqual(outputs.strings, ['foo'])



import dataclasses
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Convert targets or predictions of object detection to multilabel.

    Input detections is a list of N tensors, each tensor is objects of an image, with each row containing class_id, (score), bbox coordinates.
    If score is provided, the maximum score of each class in an image is used as the multi-label logit for that class.
    Output is a tensor of shape (N, num_classes) wich each row being multi-hot vectors or logits (if score is provided).
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        detections: typing.List[torch.Tensor]
        num_classes: int

    @dataclasses.dataclass
    class Outputs:
        results: torch.Tensor

    def execute(self, inputs):
        detections = inputs.detections
        has_score = (torch.as_tensor(detections[0]).shape[1] == 6)
        results = torch.zeros((len(detections), inputs.num_classes))

        if has_score:
            for i, img_dts in enumerate(detections):
                dts = torch.as_tensor(img_dts)
                if dts.nelement() > 0:
                    sorted_dts = dts[dts[:, 1].sort().indices]  # for each class, the maximum score is used if there are multiple boxes.
                    results[i, sorted_dts[:, 0].to(int)] = sorted_dts[:, 1]
        else:
            for i, img_dts in enumerate(detections):
                dts = torch.as_tensor(img_dts)
                if dts.nelement() > 0:
                    results[i, dts[:, 0].to(int)] = 1
                results = results.to(int)

        return self.Outputs(results)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import pathlib
import pickle
import uuid
import torch.utils.data
import irisml.core

logger = logging.getLogger()


class Task(irisml.core.TaskBase):
    """Save dataset cache on disk.

    Inputs:
        dataset (torch.utils.data.Dataset): The original dataset object. The return values must be serializable.

    Config:
        cache_dir (pathlib.Path): The path to the cache directory
    """
    VERSION = '0.1.2'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Config:
        cache_dir: pathlib.Path = pathlib.Path('.')

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    class CachedDataset(torch.utils.data.Dataset):
        def __init__(self, dataset, cache_dir):
            self._dataset = dataset
            self._cache_dir = cache_dir
            cache_dir.mkdir(parents=True, exist_ok=True)

        def __len__(self):
            return len(self._dataset)

        def __getitem__(self, index):
            cache_path = self._cache_dir / str(index)
            if cache_path.exists():
                try:
                    return pickle.loads(cache_path.read_bytes())
                except Exception:
                    pass
            item = self._dataset[index]
            cache_path.write_bytes(pickle.dumps(item))
            return item

        def __getstate__(self):
            return {'dataset': self._dataset, 'cache_dir': self._cache_dir}

        def __setstate__(self, state):
            self.__init__(**state)

        def __getattr__(self, name):
            return getattr(self._dataset, name)

    def execute(self, inputs):
        cache_dir = self.config.cache_dir / str(uuid.uuid4())
        logger.info(f"Saving dataset cache to {cache_dir}")
        return self.Outputs(Task.CachedDataset(inputs.dataset, cache_dir))

    def dry_run(self, inputs):
        return self.execute(inputs)



import json
import pathlib
import tempfile
import unittest
import torch
from irisml.tasks.load_coco_detections import Task


class TestLoadCOCODetections(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            filepath = pathlib.Path(f.name)

            # With score.
            detections = [
                {"image_id": 1, "category_id": 3, "bbox": [0, 0, 5, 5], "score": 0.4},
                {"image_id": 1, "category_id": 4, "bbox": [10, 10, 20, 20], "score": 0.5},
                {"image_id": 3, "category_id": 1, "bbox": [0, 0, 2, 2], "score": 0.2}
            ]

            filepath.write_text(json.dumps(detections))
            outputs = Task(Task.Config(filepath)).execute(Task.Inputs())
            self.assertTrue(torch.equal(outputs.detections[0], torch.tensor([[2, 0.4, 0, 0, 5, 5], [3, 0.5, 10, 10, 30, 30]])))
            self.assertTrue(torch.equal(outputs.detections[1], torch.empty((0, 6))))
            self.assertTrue(torch.equal(outputs.detections[2], torch.tensor([[0, 0.2, 0, 0, 2, 2]])))

            # Without score.
            detections = [
                {"image_id": 1, "category_id": 3, "bbox": [1, 2, 4, 5]},
                {"image_id": 1, "category_id": 4, "bbox": [10, 10, 20, 20]},
                {"image_id": 3, "category_id": 1, "bbox": [0, 0, 2, 2]}
            ]
            filepath.write_text(json.dumps(detections))
            outputs = Task(Task.Config(filepath)).execute(Task.Inputs())
            self.assertTrue(torch.equal(outputs.detections[0], torch.tensor([[2, 1, 2, 5, 7], [3, 10, 10, 30, 30]])))
            self.assertTrue(torch.equal(outputs.detections[1], torch.empty((0, 5))))
            self.assertTrue(torch.equal(outputs.detections[2], torch.tensor([[0, 0, 0, 2, 2]])))



import dataclasses
import logging
import pathlib
import time
from azure.identity import DefaultAzureCredential
from azure.storage.blob import ContainerClient
import tenacity
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Upload a directory to Azure Blob Storage.

    Config:
        container_url (str): URL of the Azure Blob Storage container to upload to.
        path (Path): Path to the directory to upload.

    Raises:
        azure.storage.exceptions if failed.

    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        container_url: str
        path: pathlib.Path

    def execute(self, inputs):
        if not self.config.path.is_dir():
            raise ValueError(f"Path {self.config.path} is not a directory")

        filepaths = sorted(f for f in self.config.path.glob('**/*') if f.is_file())
        logger.info(f"Found {len(filepaths)} files in {self.config.path}")
        container_client = ContainerClient.from_container_url(self.config.container_url, credential=DefaultAzureCredential())

        for filepath in filepaths:
            blob_name = str(filepath.relative_to(self.config.path))
            self._upload_file(container_client, blob_name, filepath)

        return self.Outputs()

    @tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_fixed(10))
    def _upload_file(self, container_client, blob_path, local_path):
        logger.info(f"Uploading {local_path} to {blob_path}")
        start = time.time()
        with open(local_path, 'rb') as f:
            container_client.upload_blob(blob_path, f, max_concurrency=8, timeout=300, overwrite=True)

        logger.info(f"Uploaded {local_path} to {blob_path} in {time.time() - start:.2f} seconds")



import dataclasses
import logging
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Floating point division.

    Inputs:
        float0 (float): The output is (float0 / float1).
        float1 (float):
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        float0: float
        float1: float

    @dataclasses.dataclass
    class Outputs:
        result: float

    def execute(self, inputs):
        result = inputs.float0 / inputs.float1
        logger.debug(f"{inputs.float0} / {inputs.float1} = {result}")
        return self.Outputs(result)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
from irisml.tasks.get_item import Task


class TestGetItem(unittest.TestCase):
    def test_simple(self):
        inputs = Task.Inputs([1, 2, 3])
        config = Task.Config(1)
        outputs = Task(config).execute(inputs)
        self.assertEqual(outputs.item, 2)

        with self.assertRaises(IndexError):
            Task(config).execute(Task.Inputs([0]))



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Run the given tasks in sequence. Each task must have an unique name."""
    VERSION = '0.1.1'
    CACHE_ENABLED = False
    RESOLVE_CONFIG_VARIABLES = False

    @dataclasses.dataclass
    class Config:
        tasks: typing.List[irisml.core.TaskDescription]

    def execute(self, inputs):
        for task_description in self.config.tasks:
            task = irisml.core.Task(task_description)
            task.load_module()
            task.execute(self.context)

        return self.Outputs()

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.emulate_fp8_quantization import Task


class TestEmulateFp8Quantization(unittest.TestCase):
    def test_allclose(self):
        inputs = (torch.rand(100) + 2 ** -6) * 10
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertTrue(torch.allclose(inputs, outputs, rtol=2 ** -2))

        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=-inputs)).tensor
        self.assertTrue(torch.allclose(-inputs, outputs, rtol=2 ** -2))

    def test_subnormal_numbers(self):
        inputs = (torch.rand(100) - 0.5) * (2 ** -6)
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertTrue(torch.equal(outputs, torch.zeros_like(outputs)))

    def test_exact_numbers(self):
        inputs = torch.tensor([0.25, 0.5, 1.0, 2.0, -0.25, -0.5, -1.0, -2.0])
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertTrue(torch.equal(outputs, inputs))

    def test_special_values(self):
        inputs = torch.tensor([0.0, 1.0, -1.0, float('inf'), float('-inf')])
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertTrue(torch.equal(outputs, inputs))

        inputs = torch.tensor([float('nan')])
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertTrue(torch.isnan(outputs).all())

    def test_out_of_range(self):
        inputs = torch.tensor([2 ** 16, -2 ** 16, 2 ** 32, -2 ** 32], dtype=torch.float32)
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertTrue(outputs.tolist(), [128, -128, 128, -128])

    def test_non_float(self):
        inputs = torch.tensor([1, 2, 3])
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(tensor=inputs))

    def test_shape(self):
        inputs = torch.rand(1, 3, 8, 8)
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor=inputs)).tensor
        self.assertEqual(outputs.shape, inputs.shape)



import dataclasses
import json
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Parse a JSON string and return a list of keys and a list of lists of ints.

    The JSON string must be a dictionary with string keys and list of ints values.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        json_string: str

    @dataclasses.dataclass
    class Outputs:
        key_names: typing.List[str]
        ints: typing.List[torch.Tensor]

    def execute(self, inputs):
        try:
            json_dict = self._decode_json(inputs.json_string)
        except Exception as e:
            logger.error(f"Could not decode JSON string: {e}")
            raise

        key_names = []
        ints = []
        for key, value in json_dict.items():
            logger.debug(f"Key: {key}, Value: {value}")
            if not isinstance(value, list) or not all(isinstance(x, int) for x in value):
                logger.warning(f"Skipping key {key} because it is not a list of ints")
                continue
            key_names.append(key)
            ints.append(torch.tensor(value))

        assert len(key_names) == len(ints)
        return self.Outputs(key_names=key_names, ints=ints)

    @staticmethod
    def _decode_json(json_str):
        start = json_str.find('{')
        if start == -1:
            raise ValueError("Could not find start of JSON string")

        try:
            decoded, _ = json.JSONDecoder().raw_decode(json_str[start:])
            return decoded
        except json.JSONDecodeError:
            pass

        try:
            # Try replacing a single quote with a double quote and vice versa.
            trans = str.maketrans('\'"', '"\'')
            decoded, _ = json.JSONDecoder().raw_decode(json_str.translate(trans)[start:])
            return decoded
        except json.JSONDecodeError:
            pass

        try:
            # Try replacing a single quote with a double quote
            decoded, _ = json.JSONDecoder().raw_decode(json_str.replace('\'', '"')[start:])
            return decoded
        except json.JSONDecodeError:
            raise



import json
import pathlib
import tempfile
import unittest
from irisml.tasks.save_str_list_jsonl import Task


class TestSaveStrListJsonl(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            temp_file = pathlib.Path(f.name)
            Task(Task.Config(path=temp_file, key_name='text')).execute(Task.Inputs(data=['a', 'b', 'c']))

            with open(temp_file) as f:
                lines = [json.loads(line) for line in f]
                self.assertEqual(lines, [{'text': 'a'}, {'text': 'b'}, {'text': 'c'}])



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Convert a string to a list of strings."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        string: str

    @dataclasses.dataclass
    class Outputs:
        strings: typing.List[str]

    def execute(self, inputs):
        return self.Outputs([inputs.string])

    def dry_run(self, inputs):
        return self.execute(inputs)



import sys
import unittest
import unittest.mock
from irisml.core import TaskDescription, TaskBase
from irisml.tasks.run_sequential import Task


class FakeTask:
    class Task(TaskBase):
        def execute(self, inputs):
            pass


class TestRunSequential(unittest.TestCase):
    def test_run(self):
        tasks = [
            {'task': 'fake_task', 'name': 'task0'},
            {'task': 'fake_task', 'name': 'task1'},
            {'task': 'fake_task', 'name': 'task2'}
        ]

        config = Task.Config(tasks=[TaskDescription.from_dict(t) for t in tasks])
        context = unittest.mock.MagicMock()

        with unittest.mock.patch.dict('sys.modules'):
            sys.modules['irisml.tasks.fake_task'] = FakeTask
            task = Task(config, context)
            task.execute(None)

        context.add_outputs.assert_any_call('task0', unittest.mock.ANY)
        context.add_outputs.assert_any_call('task1', unittest.mock.ANY)
        context.add_outputs.assert_any_call('task2', unittest.mock.ANY)



import dataclasses
import logging
import typing
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get statistics of a dataset."""
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        num_images: int = 0
        num_classes: int = 0
        num_boxes: int = 0
        dataset_type: typing.Literal['multiclass_classification', 'multilabel_classification', 'object_detection'] = None

    def execute(self, inputs):
        num_images = len(inputs.dataset)
        dataset_type = None
        num_boxes = 0
        classes_set = set()
        potential_dataset_type = set()

        for image, targets in inputs.dataset:
            if isinstance(targets, torch.Tensor):
                if not targets.shape:
                    targets = [int(targets)]
                    potential_dataset_type.add('multiclass_classification')
                elif len(targets.shape) == 1:
                    targets = [int(x) for x in targets]
                elif len(targets.shape) == 2 and targets.shape[1] == 5:
                    targets = [[int(t[0]), float(t[1]), float(t[2]), float(t[3]), float(t[4])] for t in targets]
            elif isinstance(targets, int):
                targets = [targets]
                potential_dataset_type.add('multiclass_classification')

            if len(targets) > 1:
                if isinstance(targets[0], int):
                    potential_dataset_type.add('multilabel_classification')
                elif isinstance(targets[0], list) and len(targets[0]) == 5:
                    potential_dataset_type.add('object_detection')

            if targets and isinstance(targets[0], list):
                classes_set.update(t[0] for t in targets)
            else:
                classes_set.update(targets)

        if len(potential_dataset_type) > 1:
            raise RuntimeError(f"Couldn't detect dataset types: {potential_dataset_type}")

        if potential_dataset_type:
            dataset_type = list(potential_dataset_type)[0]
        else:
            dataset_type = 'multiclass_classification'

        num_classes = len(classes_set)
        logger.info(f"Dataset type: {dataset_type}, num_classes={num_classes}, num_images={num_images}, num_boxes={num_boxes}")
        return self.Outputs(num_images=num_images, num_classes=num_classes, num_boxes=num_boxes, dataset_type=dataset_type)



import unittest
from irisml.tasks.make_prompt_for_each_string import Task


class TestMakePromptForEachString(unittest.TestCase):
    def test_simple(self):
        output = Task(Task.Config(template='What is <|placeholder|>?')).execute(Task.Inputs(strings=['a', 'b', 'c']))
        self.assertEqual(output.prompts, ['What is a?', 'What is b?', 'What is c?'])

    def test_two_strings(self):
        output = Task(Task.Config(template='What is <|placeholder|> (<|placeholder2|>)?')).execute(Task.Inputs(strings=['a', 'b', 'c'], strings2=['1', '2', '3']))
        self.assertEqual(output.prompts, ['What is a (1)?', 'What is b (2)?', 'What is c (3)?'])

    def test_invalid_config(self):
        with self.assertRaises(ValueError):
            # missing placeholder2
            Task(Task.Config(template='What is <|placeholder|>?')).execute(Task.Inputs(strings=['a', 'b', 'c'], strings2=['1', '2', '3']))

        with self.assertRaises(ValueError):
            # missing strings2
            Task(Task.Config(template='What is <|placeholder|> (<|placeholder2|>)?')).execute(Task.Inputs(strings=['a', 'b', 'c']))

        with self.assertRaises(ValueError):
            # different number of strings and strings2.
            Task(Task.Config(template='What is <|placeholder|> (<|placeholder2|>)?')).execute(Task.Inputs(strings=['a', 'b', 'c'], strings2=['1', '2']))



import unittest
import unittest.mock
from irisml.tasks.switch_pick import Task


class TestSwitchPick(unittest.TestCase):
    def test_switch_pick(self):
        test_cases = [([True, True, False], [1, 2, 3], 1), ([False, True], [1, 2], 2), ([False, False], [1, 2], None),
                      ([], [], None), ([True], [[1, 2]], [1, 2]), ([False, True], [[1, 2], [3, 4]], [3, 4])]
        for conditions, vals, expect in test_cases:
            task = Task(Task.Config())
            result = task.execute(Task.Inputs(conditions, vals)).result
            self.assertEqual(result, expect)

    def test_switch_pick_len_not_equal(self):
        task = Task(Task.Config())
        self.assertRaises(Exception, lambda: task.execute(Task.Inputs([True], [1, 2])))

    def test_default_val_works(self):
        task = Task(Task.Config())
        self.assertEqual(task.execute(Task.Inputs([False, False], [1, 2], 3)).result, 3)



import dataclasses
import json
import logging
import pathlib
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load strings from a JSON file.

    The JSON file must be a list of strings, or a dictionary with integer keys
    and string values. The strings are loaded into a list, with the index
    corresponding to the key in the dictionary.

    Config:
        filepath (Path): Path to the JSON file.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        filepath: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        strs: typing.List[str]

    def execute(self, inputs):
        data = json.loads(self.config.filepath.read_text())
        if isinstance(data, list):
            return self.Outputs(strs=data)

        if not all(k.isdigit() for k in data.keys()):
            raise ValueError('keys must be integers')

        max_key = max(int(k) for k in data.keys())
        strs = [''] * (max_key + 1)
        for k, v in data.items():
            strs[int(k)] = v
        logger.info(f"Loaded {len(data.keys())} strings to {max_key + 1} slots")
        return self.Outputs(strs=strs)

    def dry_run(self, inputs):
        return self.execute(inputs)



import random
import unittest
import torch.utils.data
from irisml.tasks.get_kfold_cross_validation_dataset import Task


class NumberDataset(torch.utils.data.Dataset):
    def __init__(self, num_samples):
        self._num_samples = num_samples

    def __len__(self):
        return self._num_samples

    def __getitem__(self, index):
        return index


class TestGetKfoldCrossValidationDataset(unittest.TestCase):
    def test_simple(self):
        dataset = NumberDataset(100)

        random.seed(0)
        outputs0 = Task(Task.Config(num_folds=3, index=0)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs0.train_dataset), 66)
        self.assertEqual(len(outputs0.val_dataset), 34)
        self.assertEqual(set(outputs0.train_dataset) | set(outputs0.val_dataset), set(range(100)))

        random.seed(0)
        outputs1 = Task(Task.Config(num_folds=3, index=1)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs1.train_dataset), 67)
        self.assertEqual(len(outputs1.val_dataset), 33)
        self.assertEqual(set(outputs1.train_dataset) | set(outputs1.val_dataset), set(range(100)))

        random.seed(0)
        outputs2 = Task(Task.Config(num_folds=3, index=2)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs2.train_dataset), 67)
        self.assertEqual(len(outputs2.val_dataset), 33)
        self.assertEqual(set(outputs2.train_dataset) | set(outputs2.val_dataset), set(range(100)))

        self.assertEqual(set(outputs0.val_dataset) | set(outputs1.val_dataset) | set(outputs2.val_dataset), set(range(100)))



import time
import unittest
from irisml.tasks.get_current_time import Task


class TestGetCurrentTime(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs())
        current_time = time.time()
        self.assertLessEqual(outputs.time, current_time)



import unittest
from irisml.tasks.compare import Task


class TestCompare(unittest.TestCase):
    TEST_CASES = [(1, 0), (0, 0), (0, 1)]

    def test_greater(self):
        self._base_test([True, False, False], False, True)

    def test_greater_or_equal(self):
        self._base_test([True, True, False], True, True)

    def test_less(self):
        self._base_test([False, False, True], False, False)

    def test_less_or_equal(self):
        self._base_test([False, True, True], True, False)

    def _base_test(self, expected, equal_allowed, greater):
        for test_case, result in zip(self.TEST_CASES, expected):
            outputs = Task(Task.Config(equal_allowed=equal_allowed, greater=greater)).execute(Task.Inputs(test_case[0], test_case[1]))
            self.assertEqual(outputs.result, result)



import unittest
import torch
from irisml.tasks.convert_detection_to_multilabel import Task


class TestConvertDetectionToMultilabel(unittest.TestCase):
    def test_simple(self):
        box = [0, 0, 0.5, 0.5]
        # class_ids: [0,1], [0,1,2]
        detections = [torch.tensor([[0, *box], [1, *box], [0, *box]]),
                      torch.tensor([[1, *box], [0, *box], [2, *box]])]
        result = Task(Task.Config()).execute(Task.Inputs(detections, num_classes=3))
        self.assertTrue(torch.equal(result.results, torch.tensor([[1, 1, 0], [1, 1, 1]])))

    def test_logits(self):
        box = [0, 0, 0.5, 0.5]
        detections_with_score = [torch.tensor([[0, 0.1, *box], [1, 0.9, *box], [0, 0.3, *box]]),
                                 torch.tensor([[1, 0.8, *box], [0, 0.3, *box], [2, 0.6, *box]])]
        result = Task(Task.Config()).execute(Task.Inputs(detections_with_score, num_classes=3))
        self.assertTrue(torch.equal(result.results, torch.tensor([[0.3, 0.9, 0.0], [0.3, 0.8, 0.6]])))



import contextlib
import os
import pathlib
import pickle
import tempfile
import unittest
import PIL.Image
import torch.utils.data
from irisml.tasks.make_cached_dataset import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]

    def get_targets(self, index):
        return self._data[index][1]


@contextlib.contextmanager
def with_chdir(new_cwd):
    new_cwd.mkdir(parents=True, exist_ok=True)
    old_cwd = os.getcwd()
    os.chdir(new_cwd)
    try:
        yield
    finally:
        os.chdir(old_cwd)


class TestMakeCachedDataset(unittest.TestCase):
    def test_classification_multiclass(self):
        dataset = FakeDataset([(PIL.Image.new('RGB', (32, 32)), torch.tensor(1)), (PIL.Image.new('RGB', (32, 32)), torch.tensor(2))])
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            outputs = Task(Task.Config(temp_dir)).execute(Task.Inputs(dataset))

            for original, cached in zip(dataset, outputs.dataset):
                self.assertEqual(original, cached)

    def test_optional_dataset_attr(self):
        dataset = FakeDataset([(PIL.Image.new('RGB', (32, 32)), torch.tensor(1)), (PIL.Image.new('RGB', (32, 32)), torch.tensor(2))])
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            outputs = Task(Task.Config(temp_dir)).execute(Task.Inputs(dataset))

            self.assertEqual(outputs.dataset.get_targets(0), torch.tensor(1))

    def test_cache(self):
        dataset = FakeDataset([(PIL.Image.new('RGB', (32, 32)), torch.tensor(1)), (PIL.Image.new('RGB', (32, 32)), torch.tensor(2))])
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            with with_chdir(temp_dir / 'old_dir'):
                outputs = Task(Task.Config(pathlib.Path('.'))).execute(Task.Inputs(dataset))
                serialized_dataset = pickle.dumps(outputs.dataset)

            with with_chdir(temp_dir / 'new_dir'):
                deserialized_dataset = pickle.loads(serialized_dataset)
                for original, cached in zip(dataset, deserialized_dataset):
                    self.assertEqual(original, cached)



import io
import unittest
import torch
from irisml.tasks.deserialize_tensor import Task


class TestDeserializeTensor(unittest.TestCase):
    def test_simple(self):
        tensor = torch.rand(3, 4, 5)
        bytes_io = io.BytesIO()
        torch.save(tensor, bytes_io)
        serialized = bytes_io.getvalue()

        outputs = Task(Task.Config()).execute(Task.Inputs(serialized))
        self.assertTrue(torch.equal(outputs.tensor, tensor))



import dataclasses
import logging
import pathlib
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Reads a file and returns its contents as bytes."""
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        data: bytes

    def execute(self, inputs):
        if not self.config.path.exists():
            raise ValueError(f"File not found: {self.config.path}")

        data = self.config.path.read_bytes()
        logger.info(f"Read {len(data)} bytes from {self.config.path}")
        return self.Outputs(data)

    def dry_run(self, inputs):
        return self.execute(inputs)



import pathlib
import unittest
from irisml.tasks.join_filepath import Task


class TestJoinFilepath(unittest.TestCase):
    def test_simple(self):
        config = Task.Config(pathlib.Path('parent_dir'), 'my_file.txt')
        outputs = Task(config).execute(Task.Inputs())
        self.assertEqual(outputs.path, pathlib.Path('parent_dir/my_file.txt'))

    def test_abs_path(self):
        config = Task.Config(pathlib.Path('/parent_dir'), 'my_file.txt')
        outputs = Task(config).execute(Task.Inputs())
        self.assertEqual(outputs.path, pathlib.Path('/parent_dir/my_file.txt'))



import unittest
from irisml.core import Context
from irisml.tasks.assertion import Task


class TestAssert(unittest.TestCase):
    def test_simple_assert(self):
        self._assert_pass(100, assert_equal=100)
        self._assert_pass('100', assert_equal='100')
        self._assert_fail(100, assert_equal='100')
        self._assert_pass(1.2345, assert_equal=1.2345)

        self._assert_pass([1, 2, 3], assert_equal=[1, 2, 3])
        self._assert_fail([3, 4], assert_equal=[1, 2])

        self._assert_pass(100, assert_not_equal=105)

        self._assert_pass(3, assert_almost_equal=3)
        self._assert_pass(3.14, assert_almost_equal=3.15)
        self._assert_fail(3.14, assert_almost_equal=3.24)
        self._assert_pass(3.14, assert_not_almost_equal=3.24)
        self._assert_pass(3.14, assert_almost_equal=3.24, delta=0.2)

        self._assert_pass(123, assert_greater=100)
        self._assert_fail(123, assert_greater=1000.5)
        self._assert_fail(123, assert_greater=123)

        self._assert_pass(123, assert_greater_equal=1)
        self._assert_pass(123, assert_greater_equal=123)
        self._assert_fail(123, assert_greater_equal=153)

        self._assert_pass(100, assert_less=105)
        self._assert_fail(100, assert_less=49)
        self._assert_fail(100, assert_less=100)

        self._assert_pass(100, assert_less_equal=200)
        self._assert_fail(100, assert_less_equal=10)
        self._assert_pass(100, assert_less_equal=100)

    def test_invalid_assert(self):
        with self.assertRaises(ValueError):
            self._execute(50, assert_almost_equal='100')

        with self.assertRaises(ValueError):
            self._execute(50, assert_greater='100')

        with self.assertRaises(ValueError):
            self._execute(50, assert_less='100')

    def _execute(self, value, **kwargs):
        inputs = Task.Inputs(value)
        config = Task.Config(**kwargs)
        task = Task(config, Context())
        task.execute(inputs)

    def _assert_pass(self, value, **kwargs):
        self._execute(value, **kwargs)

    def _assert_fail(self, value, **kwargs):
        with self.assertRaises(AssertionError):
            self._execute(value, **kwargs)
